/* Mptn -- regexp-like pattern matcher suited for morphological analysis
 * Copyright (C) 2000  George Bronnikov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
  The main control structure.

  $Id: mptn-control.c 1.3 Sun, 23 Jan 2000 20:56:51 -0600 goga $
*/

#include "mptn.h"
#include "mptn-chrset.h"
#include "mptn-private.h"

/* The global Mptn_control, used by default */
mptn_control_t *mptn_control_global = NULL;

mptn_control_t * 
mptn_control_new ()
{
  mptn_control_t *res;

  res = g_malloc (sizeof (mptn_control_t));
  res->magic = MPTN_CONTROL_MAGIC;
  res->vars_htable = g_hash_table_new (g_direct_hash, g_direct_equal);
  res->matchers_htable = g_hash_table_new (g_direct_hash, g_direct_equal);
  res->matcher_param_htable = g_hash_table_new (g_direct_hash, g_direct_equal);

  return res;
}

/* Free an Mptn_control_t */
void
mptn_control_free (mptn_control_t *control)
{
  MPTN_CONTROL_CHECK (control);

  g_hash_table_foreach (control->vars_htable, (GHFunc) mptn_free, NULL);
  g_hash_table_destroy (control->vars_htable);

  /* No free operation for matchers */
  g_hash_table_destroy (control->matchers_htable);

  g_free (control);
}

/* Restrict the possible values of a variable */
void
mptn_var_restrict (mptn_control_t *control, GQuark vcode, mptn_t *mptn)
{
  mptn_t *oldmptn;

  g_return_if_fail (vcode != 0);
  MPTN_CHECK (mptn);

  control = mptn_control_normalize (control);
  
  oldmptn = g_hash_table_lookup (control->vars_htable, (gpointer) vcode);
  if (oldmptn)
    mptn_free (oldmptn);

  mptn_refinc (mptn);
  g_hash_table_insert (control->vars_htable, (gpointer) vcode, mptn);
}

/*
  The same, but the expression is given as a string.
*/
void 
mptn_var_restrict_str (mptn_control_t *control, GQuark vcode, chr *expr)
{
  mptn_t *mptn;

  g_return_if_fail (vcode != 0);
  g_return_if_fail (expr != NULL);

  control = mptn_control_normalize (control);

  mptn = mptn_parse (control, expr);

  if (!mptn)
    {
      g_warning ("Could not compile expression %s", expr);
      return;
    }

  mptn_var_restrict (control, vcode, mptn);
  mptn_free (mptn);
}

/*
  Assign a matcher procedure to a matcher's name.
*/
void 
mptn_matcher_set (mptn_control_t *control, GQuark mcode, 
		  mptn_matcher_ops_t *matcher)
{
  g_return_if_fail (mcode != 0);
  g_return_if_fail (matcher != NULL);

  control = mptn_control_normalize (control);

  g_hash_table_insert (control->matchers_htable, (gpointer) mcode, matcher);
}

/*
  Assign the initialization parameter (data) to the matcher's name.
*/
void 
mptn_matcher_param_set (mptn_control_t *control, GQuark mcode,
			gpointer data)
{
  g_return_if_fail (mcode != 0);

  control = mptn_control_normalize (control);

  g_hash_table_insert (control->matcher_param_htable, (gpointer) mcode, data);
}


/* ---- From this point on, functions which are not publicly visible */

/* Make sure control refers to something */
mptn_control_t *
mptn_control_normalize (mptn_control_t *control)
{
  if (control)
    MPTN_CONTROL_CHECK (control);
  else /* !control */
    {
      if (!mptn_control_global)
	mptn_control_global = mptn_control_new ();
      control = mptn_control_global;
    }
  return control;
}

/* Get the mptn corresponding to a variable */
mptn_t *
mptn_control_get_var_ptn (mptn_control_t *control, GQuark vcode)
{
  MPTN_CONTROL_CHECK (control);
  g_return_val_if_fail (vcode, NULL);

  return g_hash_table_lookup (control->vars_htable, (gpointer) vcode);
}

/* Get the matcher function corresponding to a matcher's name */
mptn_matcher_ops_t *
mptn_control_get_matcher (mptn_control_t *control, GQuark mcode)
{
  g_return_val_if_fail (mcode != 0, NULL);
  control = mptn_control_normalize (control);

  return  g_hash_table_lookup (control->matchers_htable, (gpointer) mcode);
}

/* Get the matcher parameter */
gpointer 
mptn_control_get_matcher_param (mptn_control_t *control, GQuark mcode)
{
  g_return_val_if_fail (mcode != 0, NULL);
  control = mptn_control_normalize (control);

  return  g_hash_table_lookup (control->matcher_param_htable, (gpointer) mcode);
}
