/* Mptn -- regexp-like pattern matcher suited for morphological analysis
 * Copyright (C) 2000  George Bronnikov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
  Public header.

  $Id: mptn.h 1.3 Sun, 23 Jan 2000 20:56:51 -0600 goga $
*/
#ifndef MPTN_H
#define MPTN_H

#include <string.h>
#include <glib.h>

#include <stdio.h>

#include "mptn-types.h"


/* Types */
typedef struct mptn_control 	mptn_control_t;	/* Control structure */
typedef struct mptn 		mptn_t;  	/* A compiled pattern */
typedef struct mptn_var		mptn_var_t;	/* A variable inside the pattern */
typedef struct mptn_ectl	mptn_ectl_t;	/* An exec control structure */
typedef struct mptn_matcher_ops mptn_matcher_ops_t; /* A matcher function is represented by a structure */

/* 
   The structure representing a compiled expression.
*/
struct mptn {
  guint magic;
  int refcnt;

  mptn_control_t *control;
  GArray *varnames;		/* Variable names used in this expression,
				   stored as GQuarks */
  mptn_matcher_ops_t *ops;	/* Matching operations */

  gpointer mctl;		/* What the matcher initialization returned */
  gint min_len;			/* The minimum length of the matched string */
  gint max_len;			/* the maximum length of the matched string */
  gint flags;

#define MPTN_LONGER 	0x01	/* Prefer longer matches */
#define MPTN_SHORTER	0x02	/* Prefer shorter matches */
};

/*
  A structure used to encode the value of a variable
*/

struct mptn_var {
  GQuark vcode;			/* Representing variable's name */
  chr *beg;			/* The start of a var value */
  chr *end;			/* The end of var value */
  gboolean allocated;		/* Is it allocated separately? */
  guint stage;
};

/* ---- Matcher functions ---- */

/*
  A matcher is initialized at the time the expression is compiled,
  receiving, as a string, the arguments it gets, the vars array and
  some user data.
  This function is responsible
  for filling in Mctl, Min/Maxlen and may also set Flags.
  It may also add to the contents of Varnames but does _not_ set 
  the corresponding structure member (it is set by the pattern parser,
  and only for the toplevel expression).
*/
typedef mptn_t *mptn_matcher_init_ft (mptn_control_t *control, 
				      chr *args_begin, chr *args_end,
				      GArray *varnames,
				      gpointer data);

/*
  The matcher may allocate some data in mctl, this is the place to release it.
*/
typedef void mptn_matcher_destroy_ft (gpointer mctl);

/*
  At the time the pattern is matched, Matcher is called as an iterator.
  It is passed the string segment to work on, the array of/for
  variables and the "stage" to mark its own variable assignments
  and additional data.
  The iterator start routine returns iterator data, iterator step fills the vars and
  returns whether it was successful. (when not, it frees the iterator control data).
  If the iterator ends prematurely, you need to free the Ictl.
*/
typedef gpointer mptn_matcher_iter_start_ft (mptn_t *mptn,
					     chr *begin, chr *end,
					     mptn_var_t *vars, guint stage,
					     gpointer data);
typedef gpointer mptn_matcher_iter_step_ft (gpointer mictl);	/* return from iter_start */
typedef void mptn_matcher_ictl_free_ft (gpointer mictl);

/* Substitute Vars into the pattern; return a g_malloced string.
   (Could we avoid the malloc without cluttering the interface helplessly? -- Goga)
*/
typedef chr *mptn_matcher_subst_ft (mptn_t *mptn, mptn_var_t *vars, gpointer data);

/* Dump the matcher to file Fp with Offset from left */
typedef void mptn_matcher_print_ft (mptn_t *mptn, FILE *fp, guint offset);

/*
  While the iterator is running, you may inquire at which stage we are.
*/
typedef guint mptn_matcher_stage_ft (gpointer mictl);

/* Now the structure to hold all these functions together */				     
struct mptn_matcher_ops {
  mptn_matcher_init_ft 		*init;
  mptn_matcher_destroy_ft	*destroy;
  mptn_matcher_iter_start_ft 	*iter_start;
  mptn_matcher_iter_step_ft	*iter_step;
  mptn_matcher_ictl_free_ft	*ictl_free;
  mptn_matcher_stage_ft 	*stage;
  mptn_matcher_subst_ft		*subst;
  mptn_matcher_print_ft		*print;
};

/* ---- Public interface ---- */


/* Initialize an Mptn_control_t */
mptn_control_t * mptn_control_new ();

/* Free an Mptn_control_t */
void mptn_control_free (mptn_control_t *control);

/* 
   Compile a Mptn relative to Control 
   (Control may be NULL -- means global control structure)
 */
mptn_t *mptn_parse (mptn_control_t *control, chr *str_ptn);

/*
  The matcher initializers should call this function to get a 
  semi-filled mptn_t.
  The fields initialized are:
     magic (naturally)
     refcnt (to 1)
     control
*/
mptn_t *mptn_new (mptn_control_t *control);


/* 
   Start matching an mptn against a string (end is allowed to be NULL).
   Data may be used by matchers.
   Returns a cotrol structure containing the vars values for
   the first match, or NULL on failure.
*/
mptn_ectl_t *mptn_exec_start (mptn_t *mptn, chr *begin, chr *end, gpointer data);

/*
  A variant of Mptn_ectl_start when the array for the variables
  is provided by the caller (and may be partially filled),
  and the initial stage is set.
*/
mptn_ectl_t *mptn_exec_start_with_vars (mptn_t *mptn, chr *begin, chr *end,
					mptn_var_t *vars,
					guint stage,
					gpointer data);

/*
  Get the next match.
*/
mptn_ectl_t *mptn_exec_step (mptn_ectl_t *ectl);

/*
  Get the variables' values from a Ectl.
*/
mptn_var_t *mptn_ectl_vars (mptn_ectl_t *ectl);

/*
  Get the current matching stage from a Ectl.
*/
guint mptn_ectl_stage (mptn_ectl_t *ectl);

/*
  If the matching cycle is finished prematurely,
  you need to desroy Ectl.
*/
void mptn_ectl_free (mptn_ectl_t *ectl);

/*
  Get just one match, filling Vars, which has Nvars elements.
  Returns: whether the match was successful.
*/
gboolean mptn_exec_one (mptn_t *mptn, chr* begin, chr* end,
			int nvars, mptn_var_t *vars,
			gpointer data);

/*
  Substitute the variables from Vars into the pattern described by Mptn;
  return a string allocated by g_malloc or NULL on failure.
*/
chr *mptn_subst (mptn_t *mptn, mptn_var_t *vars,
		 gpointer data);

/*
  Increase the reference count of an mptn.
*/
void mptn_refinc (mptn_t *mptn);

/*
  Free an mptn.
*/
void mptn_free (mptn_t *mptn);

/*
  Dump an mptn.
*/
void mptn_print (mptn_t *mptn, FILE *fp, guint offset);

/* 
   !!!! All variable restrictions must happen before 
   you start calling Mptn_compile! 
*/

/* 
   Restrict the values of Var to those matching Mptn 
   (variable values in Mptn don't propagate!)
*/
void mptn_var_restrict (mptn_control_t *control, GQuark vcode, mptn_t *mptn);

#define mptn_vname_restrict(control,vname,mptn)				\
	mptn_var_restrict (control, g_quark_from_string (vname), mptn)

/*
  The same, but the expression is given as a string.
*/
void mptn_var_restrict_str (mptn_control_t *control, GQuark vcode, chr *expr);

#define mptn_vname_restrict_str(control,vname,expr)				\
	mptn_var_restrict_str (control, g_quark_from_string (vname), expr)

/*
  Assign a matcher procedure to a matcher's name.
*/
void mptn_matcher_set (mptn_control_t *control, GQuark mcode, 
		       mptn_matcher_ops_t *matcher);

#define mptn_mname_set(control, mname, matcher)					\
	mptn_matcher_set (control, g_quark_from_string (mname), matcher)

/*
  Assign the initialization parameter (data) to the matcher's name.
*/
void mptn_matcher_param_set (mptn_control_t *control, GQuark mcode,
			     gpointer data);

#define mptn_mname_param_set(control,mname,data)				\
	mptn_matcher_param_set (control, g_quark_from_string (mname), data)

/*
  The following procedures are only to be used inside the matchers.
*/

/*
  Get the index of a given variable inside the variable names' array.
  The varname is passed as GQuark.
  If there is no such variable yet, a place for a new variable is created
  (this should only be done during matcher registration).
*/
guint mptn_var_get_index (GArray *arr, GQuark var);

/*
  A function you can redefine to catch errors.
*/
void mptn_parse_error (char *fmt, char *str);

#endif /* already included */
